local console = require "console"
-- added helpful nativefs library to be able to utilize folders outside of save folder
nativefs = require "nativefs"
-- Slab is Gui for texteditor
Slab = require "Slab"

OS = love.system.getOS()

ColorPickFunction = ""
ColorPickVisible = false

MenuY = 0.0
if OS == 'iOS' then
	MenuY = 20.0
end
MenuHeight = 20.0

function love.run()
	if love.load then
		love.load(love.arg.parseGameArguments(arg), arg)
	end

	-- We don't want the first frame's dt to include time taken by love.load.
	if love.timer then
		love.timer.step()
	end

	local dt = 0

	-- Main loop time.
	return function()
		-- Process events.
		if love.event then
			love.event.pump()
			for name, a,b,c,d,e,f in love.event.poll() do
				if name == "quit" then
					if not love.quit or not love.quit() then
						return a or 0
					end
				end
				xpcall(function() love.handlers[name](a,b,c,d,e,f) end, handle_error)
			end
		end

		-- Update dt, as we'll be passing it to update
		if love.timer then
			dt = love.timer.step()
		end

		-- Call update and draw
		if love.update then
			xpcall(function() love.update(dt) end, handle_error)
		end -- will pass 0 if love.timer is disabled

		if love.graphics and love.graphics.isActive() then
			love.graphics.origin()
			love.graphics.clear(love.graphics.getBackgroundColor())
			if love.draw then
				xpcall(love.draw, handle_error)
			end
			love.graphics.present()
		end

		if love.timer then
			love.timer.sleep(0.001)
		end
	end
end

function love.load()
	love.keyboard.setKeyRepeat(true)
	love.keyboard.setTextInput(true)
	love.graphics.setBackgroundColor(unpack(console.BACKGROUND_COLOR))

	-- global directory to save and read files from (iOS = sourceBaseDirectory = "Documents")
	SourceBaseDir = love.filesystem.getSourceBaseDirectory()

	-- global variables for Slab
	InputText = "Write Code Here"

	-- Slab.Initialize does many things to include changing Love Keyboard handler to reference its own OnKeyPressed() & Love textinput handler to reference its own callback TextInput() and calls LoadState() [saved Slab State];
	-- In addition, uses function (not callback) Keyboard.isPressed(key) instead of Love keyboard.isDown(key)
	-- Slab.IsKeyPressed(key) calls Keyboard.isPressed(key)
	Slab.Initialize(--[[args]] nil, --[[dontInterceptEventHandlers]] true)
	LineHeight = Slab.GetTextHeight()
	editorHighlighting() --setup initial Editor text highlighting

	-- save all MiniIDE handlers
	MiniIDE = {}
	MiniIDE.draw = love.draw
	MiniIDE.update = love.update
	MiniIDE_handlers = {}
	for name,_ in pairs(love.handlers) do  -- look up names of handler events
		-- to ensure we restore nil handlers, wrap each value in a table
		MiniIDE_handlers[name] = { love[name] }
	end
end

function editorHighlighting()
	local Colors = Slab.GetStyle()
	SyntaxHighlight = {
		-- comments
		["%-%-%[%[.-%]%]"]  =Colors.CommentColor,
		["%-%-[^\n]*"]      =Colors.CommentColor,
		-- strings
		["\".-\""]      =Colors.StringColor,
		["'.-'"]        =Colors.StringColor,
		["%[%[.-%]%]"]  =Colors.StringColor,
		-- keywords (surrounded by word boundaries)
		["%f[%w]" .. "and" .."%f[%W]"]      =Colors.KeywordColor,
		["%f[%w]" .. "end" .. "%f[%W]"]     =Colors.KeywordColor,
		["%f[%w]" .. "in" .. "%f[%W]"]      =Colors.KeywordColor,
		["%f[%w]" .. "repeat" .. "%f[%W]"]  =Colors.KeywordColor,
		["%f[%w]" .. "break" .. "%f[%W]"]   =Colors.KeywordColor,
		["%f[%w]" .. "false" .. "%f[%W]"]   =Colors.KeywordColor,
		["%f[%w]" .. "local" .. "%f[%W]"]   =Colors.KeywordColor,
		["%f[%w]" .. "return" .. "%f[%W]"]  =Colors.KeywordColor,
		["%f[%w]" .. "do" .. "%f[%W]"]      =Colors.KeywordColor,
		["%f[%w]" .. "for" .. "%f[%W]"]     =Colors.KeywordColor,
		["%f[%w]" .. "nil" .. "%f[%W]"]     =Colors.KeywordColor,
		["%f[%w]" .. "then" .. "%f[%W]"]    =Colors.KeywordColor,
		["%f[%w]" .. "else" .. "%f[%W]"]    =Colors.KeywordColor,
		["%f[%w]" .. "function" .. "%f[%W]"] =Colors.KeywordColor,
		["%f[%w]" .. "not" .. "%f[%W]"]     =Colors.KeywordColor,
		["%f[%w]" .. "true" .. "%f[%W]"]    =Colors.KeywordColor,
		["%f[%w]" .. "elseif" .. "%f[%W]"]  =Colors.KeywordColor,
		["%f[%w]" .. "if" .. "%f[%W]"]      =Colors.KeywordColor,
		["%f[%w]" .. "or" .. "%f[%W]"]      =Colors.KeywordColor,
		["%f[%w]" .. "until" .. "%f[%W]"]   =Colors.KeywordColor,
		["%f[%w]" .. "while" .. "%f[%W]"]   =Colors.KeywordColor}
end

function handle_error(err)
	Error_message = debug.traceback('Error: ' .. tostring(err), --[[stack frame]]2):gsub('\n[^\n]+$', '')
	print(Error_message)  -- save error to console
	-- reset MiniIDE state
	Slab.ResetFrame()
	-- restore all MiniIDE handlers
	love.draw = MiniIDE.draw
	love.update = MiniIDE.update
	for name,def_wrap in pairs(MiniIDE_handlers) do
		love[name] = def_wrap[1]
	end
end

function love.keypressed(key, scancode, isrepeat)
	love.keyboard.setTextInput(true)  -- magic. keeps iOS from losing textinput events after switching apps.
	Slab.OnKeyPressed(key, scancode, isrepeat)
	if not(Slab.IsAnyInputFocused()) then console.keypressed(key, scancode, isrepeat) end
end

function love.textinput(text)
	Slab.OnTextInput(text)
	if not(Slab.IsAnyInputFocused()) then console.textinput(text) end
end

function love.keyreleased(key, scancode)
	Slab.OnKeyReleased(key, scancode)
	if not(Slab.IsAnyInputFocused()) then console.keyreleased(key, scancode) end
end

function love.update(dt)
	Slab.Update(dt)

	-- in pixels
	local DisplayW, DisplayH, DisplayFlags = love.window.getMode()

	-- in density-independent units
	local EditorY = love.window.fromPixels(MenuY + MenuHeight + console.FONT:getHeight() + console.VERTICAL_MARGIN)

	-- Try to position and size window to stay above a popup soft keyboard on a fairly small tablet.
	-- Larger devices will have to resize it manually. (But it should remember its size after the first time.)
	local EditorH = DisplayH-300
	if OS == 'iOS' then
		EditorH = 300  -- tested with iPad at 1194x834
	elseif OS == 'Android' then
		EditorH = EditorH-200  -- tested with Galaxy Tab A7 Lite
	end
	EditorH = love.window.fromPixels(EditorH)
	-- in density-independent units
	local DisplayWx = love.window.fromPixels(DisplayW)

	Slab.BeginWindow("EditorWindow",{Title="Editor", AutoSizeWindow=false, X=DisplayWx/2, Y=EditorY, W=DisplayWx/2, H=EditorH, IsOpen=true, AutoSizeContent=false, NoOutline=true, ConstrainPosition=true})
	local W,H = Slab.GetWindowActiveSize()
	local ScrollBarWidth = 20

	Slab.DisableDocks({"Left","Right","Bottom"})

	xpcall(function()

		-- ensure input is smaller than window to avoid nested scrollbars
		-- with some margin because scrollbars take up width when they appear
		if Slab.Input('EditBox',{Text=InputText, MultiLine=true, MultiLineW=W-ScrollBarWidth*2, W=W-5, H=H-5, Highlight=SyntaxHighlight}) then
			InputText = Slab.GetInputText()
		end

	end,
	function(err)  -- on error
		print('could not load '..LoadFilename..': '..err)
		InputText = ''  -- eliminate any confusion regarding which file we're editing
	end)

	if Slab.BeginDialog("load dialog", {Title = "Load File"}) then
		Slab.Text("Load Filename: ")
		Slab.SameLine()
		if Slab.BeginComboBox('filePicker',{Selected = LoadFilename}) then
			for index,info in ipairs(nativefs.getDirectoryItemsInfo(SourceBaseDir)) do
				if info.type == 'file' then
					if Slab.TextSelectable(info.name) then
						LoadFilename = info.name
					end
				end
			end
			Slab.EndComboBox()
		end
		Slab.NewLine()
		if Slab.Button("Load") and LoadFilename then
			InputText = nativefs.read(SourceBaseDir.."/"..LoadFilename)
			Slab.CloseDialog()
		end
		Slab.SameLine()
		if Slab.Button("Cancel") then
			Slab.CloseDialog()
		end
		Slab.EndDialog()
	end

	if Slab.BeginDialog("save dialog", {Title = "Save File"}) then
		Slab.Text("Save Filename: ")
		Slab.SameLine()
		if Slab.Input("writeFileName",{Text=SaveFilename}) then
			SaveFilename = Slab.GetInputText()
		end
		Slab.NewLine()
		if Slab.Button("Save") and SaveFilename then
			nativefs.write(SourceBaseDir.."/"..SaveFilename, InputText)
			print("file: "..SaveFilename.." written")
			Slab.CloseDialog()
		end
		Slab.SameLine()
		if Slab.Button("Cancel") then
			Slab.CloseDialog()
		end
		Slab.EndDialog()
	end

	if Slab.BeginDialog("style dialog", {Title = "Choose Editor Style "}) then
		local Style = Slab.GetStyle()
		local Names = Style.API.GetStyleNames()
		local CurrentStyle = Style.API.GetCurrentStyleName()
		if Slab.BeginComboBox('SlabStylePicker', {Selected = CurrentStyle}) then
			for I, V in ipairs(Names) do
				if Slab.TextSelectable(V) then
					Style.API.SetStyle(V)
					editorHighlighting()
					Slab.CloseDialog()
				end
			end
			Slab.EndComboBox()
		end
	Slab.EndDialog()
	end
	Slab.EndWindow()

	-- Workaround for https://github.com/flamendless/Slab/issues/164:
	-- draw the menus after everything else.
	if Slab.BeginMainMenuBar() then
		if Slab.BeginMenu("Console") then
			if Slab.MenuItem("Increase font size") then
				console.FONT_SIZE = console.FONT_SIZE + 1
				console.FONT = love.graphics.newFont(console.FONT_SIZE)
			end
			if Slab.MenuItem("Decrease font size") then
				console.FONT_SIZE = console.FONT_SIZE - 1
				console.FONT = love.graphics.newFont(console.FONT_SIZE)
			end
			if Slab.MenuItem("Clear") then
				addText("clear")
			end
			if Slab.MenuItem("Change font color") then
				ColorPickVisible = true
				ColorPickFunction = "console"
			end
		Slab.EndMenu()
		end
		if Slab.BeginMenu("Editor") then
			if Slab.MenuItem("Load") then
				Slab.OpenDialog("load dialog")
			end
			if Slab.MenuItem("Save") then
				Slab.OpenDialog("save dialog")
			end
			if Slab.MenuItem("Run") then
				addText(InputText)
			end
			if Slab.MenuItem("Copy highlight") then
				love.system.setClipboardText(Slab.GetSelection("EditorWindow.EditBox"))
			end
			if Slab.MenuItem("Paste at cursor") then
				InputText = Slab.InsertText("EditorWindow.EditBox", love.system.getClipboardText())
			end
			if Slab.MenuItem("Clear") then
				InputText = ""
			end
			if Slab.MenuItem("Change color scheme") then
				Slab.OpenDialog("style dialog")
			end
		Slab.EndMenu()
		end
		SnippetMenu()
		-- Android doesn't autodetect orientation,
		-- so provide an explicit option to toggle between landscape and portrait mode.
		if OS == 'Android' then
			if Slab.BeginMenu("Rotate device") then
				if Slab.MenuItem("Rotate device") then
					DisplayW, DisplayH = DisplayH, DisplayW
					love.window.setMode(DisplayW, DisplayH, DisplayFlags)
				end
			Slab.EndMenu()
			end
		end
	Slab.EndMainMenuBar()
	end

	if ColorPickVisible then
		Result = Slab.ColorPicker({Color = {1.0, 0.0, 0.0, 1.0}, X=love.graphics.getWidth()/6, Y = EditorY })
		if Result.Button == -1 then
			ColorPickVisible = false
		elseif Result.Button == 1 and ColorPickFunction == "console" then
			console.TEXT_COLOR = Result.Color
			ColorPickVisible = false
		elseif Result.Button == 1 and ColorPickFunction == "editor" then
			local r,g,b,a = unpack(Result.Color)
			InputText = Slab.InsertText("EditorWindow.EditBox", "{"..tostring(r)..", "..tostring(g)..", "..tostring(b)..", "..tostring(a).."}")
			ColorPickVisible = false
		end
	end
end

function love.draw()
	console.draw()  -- draws all user code if it doesn't override love.draw
	Slab.Draw()
end

function addText(text)
	console.addHistory(text)
	console.execute(text, "editor")
end

function SnippetMenu()
	if Slab.BeginMenu("Insert snippet") then
		if Slab.MenuItem("pick a color") then
			ColorPickVisible = true
			ColorPickFunction = "editor"
		end
		if Slab.MenuItem("line") then
			InputText = Slab.InsertText("EditorWindow.EditBox", 'love.graphics.line(')
		end
		if Slab.MenuItem("circle") then
			InputText = Slab.InsertText("EditorWindow.EditBox", 'love.graphics.circle(')
		end
		if Slab.MenuItem("setColor") then
			InputText = Slab.InsertText("EditorWindow.EditBox", 'love.graphics.setColor(')
		end
		if Slab.MenuItem("print") then
			InputText = Slab.InsertText("EditorWindow.EditBox", 'love.graphics.print(')
		end
		if Slab.MenuItem("love.draw") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.draw()\n\nend\n')
		end
		if Slab.MenuItem("love.update") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.update(dt)\n\nend\n')
		end
		if Slab.MenuItem("love.keypressed") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.keypressed(key, scancode, isRepeat)\n\nend\n')
		end
		if Slab.MenuItem("love.textinput") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.textinput(text)\n\nend\n')
		end
		if Slab.MenuItem("love.keyreleased") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.keyreleased(key, scancode)\n\nend\n')
		end
		if Slab.MenuItem("love.mousepressed") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.mousepressed(x,y, button, istouch, presses)\n\nend\n')
		end
		if Slab.MenuItem("love.mousereleased") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.mousereleased(x,y, button, istouch, presses)\n\nend\n')
		end
		if Slab.MenuItem("love.wheelmoved") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.wheelmoved(x,y)\n\nend\n')
		end
		if Slab.MenuItem("love.touchpressed") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.touchpressed(id, x,y, dx,dy, pressure)\n\nend\n')
		end
		if Slab.MenuItem("love.touchreleased") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.touchreleased(id, x,y, dx,dy, pressure)\n\nend\n')
		end
		if Slab.MenuItem("love.touchmoved") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\nfunction love.touchmoved(id, x,y, dx,dy, pressure)\n\nend\n')
		end
		if Slab.MenuItem("background color") then
			InputText = Slab.InsertText("EditorWindow.EditBox", '\n  love.graphics.setBackgroundColor(1,1,1)\n')
		end
		if Slab.MenuItem("example: full-screen app") then
			InputText = InputText..[[

-- Example script using Slab GUI library that you can modify to your needs.
-- Puts a single button on screen to return to MiniIDE.
-- There are many more GUI elements you can add to your apps:
--   https://github.com/flamendless/Slab/wiki

function love.update(dt)
	Slab.Update(dt)
	Slab.BeginWindow('window 1', {X=love.graphics.getWidth()-200,Y=100, ConstrainPosition=true})
	if Slab.Button('back to MiniIDE') then
		escape()
	end
	Slab.EndWindow()
end

function love.draw()
	Slab.Draw()
	love.graphics.setColor({1,0,0})
	love.graphics.line(100,100, 200,300)
	love.graphics.setColor({0,1,0})
	love.graphics.circle('fill', 150,150, 60)
end

function escape()
	error('StopScript')
end

-- Set LOVE2D callbacks to Slab callbacks when use Slab GUI:
function love.keypressed(key, scancode, isrepeat)
	Slab.OnKeyPressed(key, scancode, isrepeat)
end

function love.textinput(text)
	Slab.OnTextInput(text)
end

function love.keyreleased(key, scancode)
	Slab.OnKeyReleased(key, scancode)
end

function love.mousepressed(x, y, button, istouch, presses)
	Slab.OnMousePressed(x, y, button, istouch, presses)
end

function love.mousereleased(x, y, button, istouch, presses)
	Slab.OnMouseReleased(x, y, button, istouch, presses)
end

function love.mousemoved(x, y, dx, dy, istouch)
	Slab.OnMouseMoved(x, y, dx, dy, istouch)
end

function love.wheelmoved(x, y)
	Slab.OnWheelMoved(x, y)
end

function love.touchpressed(id, x,y, dx,dy, pressure)
	-- no Slab handler
end

function love.touchreleased(id, x,y, dx,dy, pressure)
	-- no Slab handler
end

function love.touchmoved(id, x,y, dx,dy, pressure)
	-- no Slab handler
end

function love.resize(w,h)
	-- no Slab handler
end
]]
		end
		if Slab.MenuItem("example: return to MiniIDE") then
			InputText = InputText..[[

-- call this from a button or hotkey to return to MiniIDE
function escape()
	error('StopScript')
end
]]
		end
	Slab.EndMenu()
	end
end

function love.quit()
	Slab.OnQuit()
end

function love.wheelmoved(x, y)
	Slab.OnWheelMoved(x, y)
end

function love.mousemoved(x, y, dx, dy, istouch)
	Slab.OnMouseMoved(x, y, dx, dy, istouch)
end

function love.mousepressed(x, y, button, istouch, presses)
	Slab.OnMousePressed(x, y, button, istouch, presses)
	if y > MenuY and y < MenuY + MenuHeight + console.FONT:getHeight() + console.VERTICAL_MARGIN then
		love.keyboard.setTextInput(true)
	end
end

function love.mousereleased(x, y, button, istouch, presses)
	Slab.OnMouseReleased(x, y, button, istouch, presses)
end
